"""
Bioinformatics
==============

"""

NAME = "Bioinformatics"

ICON = "icons/Category-Bioinformatics.svg"

BACKGROUND = "light-grass"

intersphinx = (
    ("{DEVELOP_ROOT}/docs/build/html/", None),
    ("http://orange-bioinformatics.readthedocs.org/en/latest/", None)
)
